/******************************************************************
 * Author: Steve Wolfman
 * File: mazerunner.c
 *
 *  Various functions for working with mazes, designed to
 * demonstrate file I/O, multi-D arrays, and recursion.
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define MAX_HEIGHT 100
#define MAX_WIDTH  100

#define WALL          '+'
#define EXIT          'X'
#define ENTRANCE      '*'
#define EMPTY         ' '
#define PATH          '.'

typedef struct {
	char type;     /* Can be EMPTY, WALL, ENTRANCE, or EXIT. */
	int  bVisited; /* Have we visited this as we searched?   */
	int  bOnPath;  /* Is this on the final path to the exit? */
} MazeCell;

typedef struct {
	int height;   /* The dimensions  */
	int width;    /* of the maze.    */
	
	int enterRow, enterCol;  /* The indexes of the entrance.   */

	MazeCell cells[MAX_HEIGHT][MAX_WIDTH]; /* The maze itself. */
} Maze;

/*************************************************
* readMaze
*
*   Given a filename to read from, reads
*   a maze into the maze struct supplied.
*
*     filename:   the name of a maze file
*     maze:       pointer to a maze structure
*
*precondition: filename names a valid maze
*              file (see below). maze points
*              to a maze structure.
*postcondition: maze has been updated to
*               describe the given maze.
*
* A maze file starts with the height and width
* of the maze on one line and is written with 
* one row of the maze per line. A ' ' is empty,
* a '*' is entrance, a 'X' is exit. Anything else
* is a wall.
*
* For example, strip the leading "* " from the
* next lines and you get a maze that looks
* somewhat like an O with the entrance in the 
* upper left and the exit in the lower right.
*
* 7 5
* +---+
* |*  |
* | ^ |
* | | |
* | v |
* |  X|
* +---+
*
* Note that there may be only one entrance; 
* however, there may be many exits. Also note
* that the first line must be exactly of the 
* form number space number newline, and the 
* last line should have a newline.
*
*************************************************/
void readMaze(char filename[], Maze * maze)
{
	FILE * fin;              /* The input file.                */
	char next;               /* The next character to process. */
	int status;              
	int col = 0, row = 0;    /* Current row/col in the maze.   */
	int bEnter = 0;          /* Have we seen an entrance?      */

	/* Open the input file for reading. Check for failure. */
	/* Note: NULL is probably new to you. In this context, */
	/* It is just another "sentinel" status, warning of an */
	/* error opening the file.                             */
	fin = fopen(filename, "r");
	if (fin == NULL) {
		printf("ERROR: problem opening file %s.\n", filename);
		exit(EXIT_FAILURE);
	}

	/* Read the dimensions (and handle errors). */
	status = fscanf(fin, "%d %d", &(maze->height), &(maze->width));
	if (status < 2) {
		printf("ERROR: bad height/width.\n");
		exit(EXIT_FAILURE);
	}
	else if (maze->height > MAX_HEIGHT || 
			 maze->width > MAX_WIDTH) {
		printf("ERROR: dimensions of %d by %d are too large.\n"
			   "Maze should be no larger than %d by %d.",
			   maze->height, maze->width, MAX_HEIGHT, MAX_WIDTH);
		exit(EXIT_FAILURE);
	}

	/* Strip the newline from the end of the line. */
	status = fscanf(fin, "%c", &next);
	if (status < 1 || next != '\n') {
		printf("ERROR: bad format for height/width line of maze.\n");
		exit(EXIT_FAILURE);
	}


	/* Read in each cell of the maze (all height by width of them). */
	for (row = 0; row < maze->height; row++) {
		for (col = 0; col < maze->width; col++) {
			status = fscanf(fin, "%c", &next);
			if (status < 1) {
				printf("ERROR: bad maze cell entry at (%d, %d).\n",
					   row, col);
				exit(EXIT_FAILURE);
			}

			/* Figure out what type of thing this cell is. */
			switch (next) {
			case EMPTY:
			case EXIT:
				maze->cells[row][col].type     = next;
				break;
			case ENTRANCE:
				if (bEnter) {
					printf("ERROR: more than one entrance!\n");
					exit(EXIT_FAILURE);					
				}
				bEnter = 1;
				maze->cells[row][col].type     = ENTRANCE;
				maze->enterCol = col;
				maze->enterRow = row;
				break;
			default:
				maze->cells[row][col].type     = WALL;
			}

			/* Initialize the other fields of the cell. */
			maze->cells[row][col].bVisited = 0;
			maze->cells[row][col].bOnPath  = 0;
		}

		/* Eat the newline at the end of the line. */
		status = fscanf(fin, "%c", &next);
		if (status < 1 || next != '\n') {
			printf("ERROR: bad row termination in maze at row %d.\n", row);
			exit(EXIT_FAILURE);
		}
	}

	/* At this point, the file should be finished: */
	if (fscanf(fin, "%c", &next) != EOF) {
		printf("WARNING: more data left in file after maze read.\n");
	}

	fclose(fin);
}


/**************************************************
* writeMaze
*
*   Given a filename to write to, writes out
*   a maze into the file. Note: if "bPrintPath"
*   is true, the resulting maze will not quite
*   be valid for reading.
*
*     filename:   the name of a file
*     maze:       a maze structure
*     bPrintPath: should the path be printed?
*
*precondition: filename available for writing. 
*              maze is a well-formed maze.
*postcondition: filename contains a file repr of
*               the maze (see readMaze for more 
*               info). However, if bPrintPath is
*               true, path will be marked with '.'
**************************************************/
void writeMaze(char * filename, Maze maze, int bPrintPath)
{
	FILE * fout;   /* Output file.           */
	int row, col;  /* Indexes into the maze. */

	/* Open the file and check for problems. */
	fout = fopen(filename, "w");
	if (fout == NULL) {
		printf("ERROR: cannot open file %s for writing.\n", filename);
		exit(EXIT_FAILURE);
	}

	/* Write the file dimensions. */
	fprintf(fout, "%d %d\n", maze.height, maze.width);

	/* Loop over the maze, printing out each element. */
	for (row = 0; row < maze.height; row++) {
		for (col = 0; col < maze.width; col++) {
			switch(maze.cells[row][col].type) {
			case WALL:
				fprintf(fout, "%c", WALL);
				break;
			case EMPTY:
				/* If we're printing the path and this is a path cell, */
				/* print a path icon. Otherwise, just empty space.     */
				if (bPrintPath && maze.cells[row][col].bOnPath) {
					fprintf(fout, "%c", PATH);
				}
				else {
					fprintf(fout, "%c", EMPTY);
				}
				break;
			case ENTRANCE:
				fprintf(fout, "%c", ENTRANCE);
				break;
			case EXIT:
				fprintf(fout, "%c", EXIT);
				break;
			default:
				printf("ERROR: unknown cell type '%c' at (%d, %d).\n",
					   maze.cells[row][col].type, row, col);
				exit(EXIT_FAILURE);
			}
		}
		/* Finish the line. */
		fprintf(fout, "\n");
	}

	/* Close the file. */
	fclose(fout);
}


/*****************************************************************
*  searchMazeHelper
*
*  NOT TO BE CALLED DIRECTLY (call searchMaze instead)
*    Given a maze, searches the maze for a path from the 
*    given cell to any exit. If found, marks that with the bOnPath
*    entries of the cells. Returns the length of the path found
*    or -1 if no path is found. (Paths use cardinal directions;
*    no diagonals).
*
*   maze:     a maze to search (a pointer for efficiency reasons)
*   row, col: indexes to start search from
*
*precondition: before first call, maze should be initialized
*              so all bVisited vars are false and all bOnPath vars
*              are false. row and col should be the enterRow and
*              enterCol for full paths.
*postcondition: if a path to the exit exists, it is marked, and
*               the length of the discovered path is returned. 
*               Otherwise, returns -1 and no path is marked.
******************************************************************/
int searchMazeHelper(Maze * maze, int row, int col)
{
	int result;  /* recursive result from sub-calls. */

	/* Error-checking base case: out of bounds. */
	if (row < 0 || row >= maze->height ||
		col < 0 || col >= maze->width) {
		return -1;
	}
	/* Successful base case: found the exit. */
	else if (maze->cells[row][col].type == EXIT) {
		/* So far the path is zero steps long. */
		return 0;
	}
	/* Failure base case: found a place we'd already been OR
	   found a wall. */
	else if (maze->cells[row][col].bVisited ||
			 maze->cells[row][col].type == WALL) {
		return -1;
	}
	/* General case: more searching to do. */
	else {
		/* Set up this square for further searching. */
		maze->cells[row][col].bOnPath  = 1;
		maze->cells[row][col].bVisited = 1;

		/* Search north. */
		result = searchMazeHelper(maze, row - 1, col);
		if (result >= 0) {
			return result + 1;
		}

		/* Search east. */
		result = searchMazeHelper(maze, row, col + 1);
		if (result >= 0) {
			return result + 1;
		}

		/* Search south. */
		result = searchMazeHelper(maze, row + 1, col);
		if (result >= 0) {
			return result + 1;
		}

		/* Search west. */
		result = searchMazeHelper(maze, row, col - 1);
		if (result >= 0) {
			return result + 1;
		}

		/* No path found. */
		maze->cells[row][col].bOnPath  = 0;
		return -1;
	}

	/* Should never reach here. */
}

/*****************************************************************
*  searchMaze
*
*    Given a maze, searches the maze for a path from the 
*    entrance to any exit. If found, marks that with the bOnPath
*    entries of the cells. Returns the length of the path found
*    or -1 if no path is found. (Paths use cardinal directions;
*    no diagonals).
*
*   maze:     a maze to search (a pointer for efficiency reasons)
*
*precondition:  maze must be valid maze (have an entrance, etc)
*postcondition: if a path to the exit exists, it is marked, and
*               the length of the discovered path is returned. 
*               Otherwise, returns -1 and no path is marked.
******************************************************************/
int searchMaze(Maze * maze)
{
	int row, col;

	/* Initialize the maze. */
	for (row = 0; row < maze->height; row++) {
		for (col = 0; col < maze->width; col++) {
			maze->cells[row][col].bOnPath  = 0;
			maze->cells[row][col].bVisited = 0;
		}
	}

	/* Call helper with entrance as start point. */
	return searchMazeHelper(maze, maze->enterRow, maze->enterCol);
}


/* Main program just to demonstrate what existing functions can do. */
int main(void)
{
	Maze maze;
	char input[101];  /* File   */
	char output[101]; /* names. */

	/* Get the input maze. */
	printf("Please enter a file to read from: ");
	scanf("%s", input);

	readMaze(input, &maze);

	/* Search it. */
	searchMaze(&maze);

	/* Write the output maze, noting the path. */
	printf("Please enter a file to write to: ");
	scanf("%s", output);

	writeMaze(output, maze, 1);
}
